#!/usr/sbin/dtrace -Zs
/*
 * js_objgcx.d - measure JavaScript Object GC using DTrace, with extra info.
 *               Written for the Solaris Nevada DTrace JavaScript provider.
 *
 * 14-Sep-2006, ver 0.80
 *
 * USAGE: js_objgcx.d		# hit Ctrl-C to end
 *
 * FIELDS:
 *              FILE		Filename that contained the function
 *              LINENO		File line number where object was created
 *              CLASS		Class to which this new object belongs
 *              TOTAL		Object entropy (positive == uncollected)
 *
 * This script provides information on which objects are not being garbage
 * collected, an issue which causes the browser to steadily leak memory.
 * We trace object creation (+1) and destruction (-1), and provide a
 * summary each 5 seconds of the running tally of the object class and
 * originating filename. If over the period of several minutes an object
 * type is still steadily increasing, then that would be of interest.
 * Be patient, depending on the rate of object creation, it can take over
 * ten minutes for garbage collect to kick in.
 *
 * ALERT: if you see "dynamic variable drops", these will cause "<missed>"
 * entries of a negative value. This may happen if you have a very busy
 * application for DTrace to monitor. To remedy this, increase the dynvarsize
 * tunable below.
 *
 * NOTE: there are other Things that GC handles, other than Objects; extra
 * probes can be added to trace them, should the need arise.
 */
/*
 * Copyright 2006 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
#pragma ident	"%Z%%M%	%I%	%E% SMI"

#pragma D option quiet
#pragma D option dynvarsize=64m
/* max entire URL, due to copyinstr, */
#pragma D option strsize=64

dtrace:::BEGIN
{
	printf("Tracing... Hit Ctrl-C to end.\n");
}

javascript*:::object-create
/arg2/
{
	this->file = basename(copyinstr(arg0));
	@objs[this->file, (int)arg3, copyinstr(arg1)] = sum(1);
	filename[arg2] = this->file;
	lineno[arg2] = (int)arg3;
}

javascript*:::object-finalize
/filename[arg2] == NULL/
{
	/* these are caused by dynvardrops. increase dynvarsize */
	@objs["<missed>", -1, copyinstr(arg1)] = sum(-1);
}

javascript*:::object-finalize
/filename[arg2] != NULL/
{
	@objs[filename[arg2], lineno[arg2], copyinstr(arg1)] = sum(-1);
	filename[arg2] = 0;
	lineno[arg2] = 0;
}

profile:::tick-5sec,
dtrace:::END
{
	printf("\n %-24s %-7s %-36s %8s\n", "FILE", "LINENO", "CLASS", "TOTAL");
	printa(" %-24.24s %-7d %-36s %@8d\n", @objs);
}
